from typing import Optional

import discord
from cpl_core.configuration import ConfigurationABC
from cpl_core.logging import LoggerABC
from cpl_discord.service import DiscordBotServiceABC

from bot_core.abc.permission_service_abc import PermissionServiceABC
from bot_data.abc.server_config_repository_abc import ServerConfigRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.technician_config_repository_abc import TechnicianConfigRepositoryABC
from bot_data.model.team_member_type_enum import TeamMemberTypeEnum


class PermissionService(PermissionServiceABC):
    def __init__(
        self,
        logger: LoggerABC,
        bot: DiscordBotServiceABC,
        config: ConfigurationABC,
        servers: ServerRepositoryABC,
        server_configs: ServerConfigRepositoryABC,
        technician_configs: TechnicianConfigRepositoryABC,
    ):
        PermissionServiceABC.__init__(self)
        self._logger = logger
        self._bot = bot
        self._config = config
        self._servers = servers
        self._server_configs = server_configs
        self._technician_configs = technician_configs

        # member_id: {team_member_type: {guild_id: bool}}
        self._cache: dict[int, dict[TeamMemberTypeEnum, dict[int, bool]]] = {}

    def reset_cache(self):
        self._cache = {}

    def get_cached_permission(
        self, member_id: int, team_member_type: TeamMemberTypeEnum, guild_id: int = None
    ) -> Optional[bool]:
        if member_id not in self._cache:
            self._cache[member_id] = {}

        if team_member_type not in self._cache[member_id]:
            self._cache[member_id][team_member_type] = {}
            return None

        if guild_id not in self._cache[member_id][team_member_type]:
            return None

        return self._cache[member_id][team_member_type][guild_id]

    def set_cached_permission(
        self, value: bool, member_id: int, team_member_type: TeamMemberTypeEnum, guild_id: int = None
    ):
        if member_id not in self._cache:
            self._cache[member_id] = {}

        if team_member_type not in self._cache[member_id]:
            self._cache[member_id][team_member_type] = {}

        self._cache[member_id][team_member_type][guild_id] = value

    def _has_member_role(self, member: discord.Member, team_member_type: TeamMemberTypeEnum) -> bool:
        if member is None or member.guild is None:
            return False

        try:
            has_permission_cached = self.get_cached_permission(member.id, team_member_type, member.guild.id)
            if has_permission_cached is not None:
                return has_permission_cached

            self._logger.debug(__name__, f"Checking is member {member.name} {team_member_type.value}")

            has_permission = True in [
                member.guild.get_role(x.role_id) in member.roles
                for x in self._server_configs.get_server_config_by_server(
                    self._servers.get_server_by_discord_id(member.guild.id).id
                ).team_role_ids
                if x.team_member_type == team_member_type
            ]
            self.set_cached_permission(has_permission, member.id, team_member_type, member.guild.id)
            return has_permission
        except Exception as e:
            self._logger.error(__name__, "Permission check failed", e)

        return False

    def is_member_admin(self, member: discord.Member) -> bool:
        return self._has_member_role(member, TeamMemberTypeEnum.admin)

    def is_member_moderator(self, member: discord.Member) -> bool:
        return self._has_member_role(member, TeamMemberTypeEnum.moderator) or self._has_member_role(
            member, TeamMemberTypeEnum.admin
        )

    def is_member_technician(self, member: discord.Member) -> bool:
        if member is None or member.guild is None:
            return False

        has_permission_cached = self.get_cached_permission(member.id, TeamMemberTypeEnum.technician)
        if has_permission_cached is not None:
            return has_permission_cached

        self._logger.debug(__name__, f"Checking is member {member.name} technician")

        try:
            has_permission = member.id in self._technician_configs.get_technician_config().technician_ids
            self.set_cached_permission(has_permission, member.id, TeamMemberTypeEnum.technician)
            return has_permission
        except Exception as e:
            self._logger.error(__name__, "Permission check failed", e)
            return False

    def is_member_technician_by_id(self, member_id: int):
        has_permission_cached = self.get_cached_permission(member_id, TeamMemberTypeEnum.technician)
        if has_permission_cached is not None:
            return has_permission_cached

        self._logger.debug(__name__, f"Checking is member {member_id} technician")

        try:
            has_permission = member_id in self._technician_configs.get_technician_config().technician_ids
            self.set_cached_permission(has_permission, member_id, TeamMemberTypeEnum.technician)
            return has_permission
        except Exception as e:
            self._logger.error(__name__, "Permission check failed", e)
            return False
