import uuid

from cpl_core.database import DatabaseSettings
from cpl_core.database.context import DatabaseContext

from bot_core.exception.service_error_code_enum import ServiceErrorCode
from bot_core.exception.service_exception import ServiceException
from bot_core.logging.database_logger import DatabaseLogger
from bot_data.mysql_pool import MySQLPool


class DBContext(DatabaseContext):
    def __init__(self, logger: DatabaseLogger):
        self._logger = logger

        DatabaseContext.__init__(self)
        self._pool: MySQLPool = None
        self._fails = 0

    def connect(self, database_settings: DatabaseSettings):
        try:
            self._logger.debug(__name__, "Connecting to database")
            self._pool = MySQLPool(database_settings)
            self._pool.execute(f"CREATE DATABASE IF NOT EXISTS `{database_settings.database}`;", commit=True)
            self._logger.info(__name__, "Connected to database")
        except Exception as e:
            self._logger.fatal(__name__, "Connecting to database failed", e)

    @property
    def cursor(self):
        return self

    def save_changes(self):
        pass

    def select(self, statement: str) -> list[tuple]:
        try:
            return self._pool.execute(statement)
        except Exception as e:
            if self._fails >= 3:
                self._logger.error(__name__, f"Database error caused by {statement}", e)
                uid = uuid.uuid4()
                raise ServiceException(
                    ServiceErrorCode.Unknown,
                    f"Query failed three times with {type(e).__name__}. Contact an admin and give them the UID: {uid}",
                )

            self._logger.error(__name__, f"Database error caused by {statement}", e)
            self._fails += 1
            try:
                self._logger.debug(__name__, "Retry select")
                return self.select(statement)
            except Exception as e:
                pass
            return []

    def execute(self, statement: str):
        return self._pool.execute(statement, commit=True)
