import glob
import os

from cpl_core.dependency_injection import ServiceProviderABC
from cpl_query.extension import List
from packaging import version

import bot
from bot_core.logging.database_logger import DatabaseLogger
from bot_data.db_context import DBContext
from bot_data.model.migration import Migration
from bot_data.model.migration_history import MigrationHistory


class MigrationService:
    def __init__(
        self,
        logger: DatabaseLogger,
        services: ServiceProviderABC,
        db: DBContext,
    ):
        self._logger = logger
        self._services = services

        self._db = db

    def _get_migration_history(self) -> List[MigrationHistory]:
        results = self._db.select(
            """
            SELECT * FROM `MigrationHistory`
        """
        )
        applied_migrations = List(MigrationHistory)
        for result in results:
            applied_migrations.add(MigrationHistory(result[0], result[1]))

        return applied_migrations

    def _migration_migrations_to_old(self, migration: MigrationHistory):
        if migration.migration_id.endswith("Migration"):
            return

        self._logger.debug(__name__, f"Migrate new migration {migration.migration_id} to old method")
        self._db.execute(migration.change_id_string(f"{migration.migration_id}Migration"))
        self._db.save_changes()

    def _migration_migrations_to_new(self, migration: MigrationHistory):
        if not migration.migration_id.endswith("Migration"):
            return

        self._logger.debug(__name__, f"Migrate old migration {migration.migration_id} to new method")
        self._db.execute(migration.change_id_string(migration.migration_id.replace("Migration", "")))
        self._db.save_changes()

    def _migrate_from_old_to_new(self):
        result = self._db.select("SHOW TABLES LIKE 'MigrationHistory'")
        if not result:
            return

        for migration in self._get_migration_history():
            if version.Version(bot.__version__) < version.Version("1.2.2"):
                self._migration_migrations_to_old(migration)
            else:
                self._migration_migrations_to_new(migration)

    def _load_scripts(self, upgrade: bool = True) -> List[Migration]:
        migrations = List(Migration)
        path = "../bot_data/scripts"

        if not os.path.exists(path):
            raise Exception("Migration path not found")

        folders = List(str, glob.glob(f"{path}/*"))
        if upgrade:
            folders = folders.order_by()
        else:
            folders = folders.order_by_descending()

        for folder in folders:
            splitted = folder.split("/")
            version_str = splitted[len(splitted) - 1]

            # upgrade do not run migrations from higher versions
            if upgrade and version.Version(version_str) > version.Version(bot.__version__):
                break
            # downgrade run migrations from higher versions
            if not upgrade and version.Version(version_str) <= version.Version(bot.__version__):
                continue

            files = List(str, os.listdir(folder)).where(lambda x: x.endswith(f"_{'up' if upgrade else 'down'}.sql"))
            if upgrade:
                files = files.order_by()
            else:
                files = files.order_by_descending()

            for file in files.to_list():
                if not file.endswith(".sql"):
                    continue

                name = str(file.split(".sql")[0])
                if name.endswith("_up"):
                    name = name.replace("_up", "")
                elif name.endswith("_down"):
                    name = name.replace("_down", "")

                name = name.split("_")[1]

                with open(f"{folder}/{file}", "r") as f:
                    script = f.read()
                    f.close()

                migrations.add(Migration(name, version_str, script))

        return migrations

    def _execute(self, migrations: List[Migration], upgrade: bool = True):
        for migration in migrations:
            active_statement = ""
            try:
                # check if table exists
                result = self._db.select("SHOW TABLES LIKE 'MigrationHistory'")
                if result:
                    # there is a table named "tableName"
                    self._logger.trace(
                        __name__,
                        f"Running SQL Command: {MigrationHistory.get_select_by_id_string(migration.name)}",
                    )
                    migration_from_db = self._db.select(MigrationHistory.get_select_by_id_string(migration.name))
                    if upgrade and migration_from_db is not None and len(migration_from_db) > 0:
                        continue
                    elif not upgrade and (migration_from_db is None or len(migration_from_db) == 0):
                        continue

                self._logger.debug(
                    __name__, f"Running {'upgrade' if upgrade else 'downgrade'} migration: {migration.name}"
                )

                for statement in migration.script.split("\n\n"):
                    if statement in ["", "\n"]:
                        continue
                    active_statement = statement
                    self._db.execute(statement + ";")

                self._db.execute(
                    MigrationHistory(migration.name).insert_string
                    if upgrade
                    else MigrationHistory(migration.name).delete_string
                )
                self._db.save_changes()
            except Exception as e:
                self._logger.fatal(
                    __name__, f"Migration failed: {migration.version}-{migration.name}\n{active_statement}", e
                )

    def migrate(self):
        self._migrate_from_old_to_new()
        self._execute(self._load_scripts())
        self._execute(self._load_scripts(False), False)
